"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Container = void 0;
const types_1 = require("./types");
const Definition_1 = require("./Definition");
const errors = require("./errors");
const assertNoCircularDependencies_1 = require("./assertNoCircularDependencies");
const ContainerArg_1 = require("./ContainerArg");
const is = require("predicates");
const randomName_1 = require("./randomName");
const debug_1 = require("./debug");
function isThenable(result) {
    return result && 'then' in result;
}
const debugCreation = debug_1.debug('creation');
const debugDefinition = debug_1.debug('definition');
class Container {
    constructor(parent) {
        this.definitions = new Map();
        this.services = new Map();
        this.middlewares = [];
        Object.defineProperty(this, 'parent', {
            value: parent,
            writable: false
        });
    }
    /**
     * Registers given service definition
     */
    registerDefinition(definition) {
        if (this.definitions.has(definition.name)) {
            throw errors.ALREADY_DEFINED(`Service "${definition.name.toString()}" already defined`);
        }
        debugDefinition(`Service ${definition.name.toString()} defined`);
        this.definitions.set(definition.name, definition);
        return this;
    }
    /**
     * Creates and registers service definition
     *
     * Returns created definition for further configuration
     */
    definition(name) {
        const definition = new Definition_1.Definition(name);
        this.registerDefinition(definition);
        return definition;
    }
    definitionWithConstructor(nameOrClazz, clazz) {
        const finalClazz = is.func(nameOrClazz) ? nameOrClazz : clazz;
        return this.definition(types_1.ServiceName.is(nameOrClazz) ? nameOrClazz : randomName_1.randomName(finalClazz.name))
            .useConstructor(finalClazz);
    }
    definitionWithFactory(nameOrFactory, factoryOrType, type) {
        const name = types_1.ServiceName.is(nameOrFactory) ? nameOrFactory : undefined;
        const factory = is.func(nameOrFactory) ? nameOrFactory : factoryOrType;
        const finalType = is.func(nameOrFactory) ? factoryOrType : type;
        const def = this.definition(name)
            .useFactory(factory);
        if (finalType) {
            def.markType(finalType);
        }
        return def;
    }
    definitionWithValue(nameOrValue, value) {
        const isNameProvided = value !== undefined && types_1.ServiceName.is(nameOrValue);
        const finalValue = isNameProvided ? value : nameOrValue;
        const name = isNameProvided ? nameOrValue : randomName_1.randomName(Object.getPrototypeOf(finalValue).constructor.name);
        return this.definition(name)
            .useValue(finalValue);
    }
    /**
     * Returns definition by given name
     */
    findByName(name) {
        if (this.definitions.has(name)) {
            return this.definitions.get(name);
        }
        if (this.parent) {
            return this.parent.findByName(name);
        }
    }
    /**
     * Returns definitions that satisfy given predicate
     */
    findByPredicate(predicate) {
        return Array.from(this.definitions.values())
            .filter(predicate)
            .concat(this.parent ? this.parent.findByPredicate(predicate) : []);
    }
    findByAnnotation(predicate, withAnnotation = false) {
        let annotations = [];
        const definitions = this.findByPredicate(s => {
            const annotation = s.annotations.find(predicate);
            if (annotation) {
                withAnnotation && annotations.push(annotation);
                return true;
            }
            return false;
        });
        if (withAnnotation) {
            return definitions.map((d) => {
                return [d, annotations.shift()];
            });
        }
        return definitions;
    }
    /**
     * Registers given middleware
     */
    addMiddleware(...middlewares) {
        this.middlewares.push(...middlewares);
        return this;
    }
    getMiddlewares() {
        return this.middlewares;
    }
    /**
     * Returns service for given name or definition
     */
    get(nameOrDefinition) {
        let definition;
        if (types_1.ServiceName.is(nameOrDefinition)) {
            definition = this.findByName(nameOrDefinition);
            if (definition === undefined) {
                return Promise.reject(errors.SERVICE_NOT_FOUND(`Service "${nameOrDefinition.toString()}" does not exist`));
            }
        }
        else {
            definition = nameOrDefinition;
        }
        if (this.services.has(definition)) {
            return this.services.get(definition);
        }
        if (!this.definitions.has(definition.name) && this.parent) {
            return this.parent.get(definition);
        }
        const promise = this.create(definition);
        this.services.set(definition, promise);
        return promise;
    }
    create(definition) {
        if (!definition.factory) {
            return Promise.reject(errors.INCOMPLETE_DEFINITION(`Missing factory for service definition "${definition.name.toString()}". Define it as constructor, factory or value`));
        }
        try {
            assertNoCircularDependencies_1.assertNoCircularDependencies(this, definition);
        }
        catch (e) {
            return Promise.reject(e);
        }
        // valid definition, time to lock it
        definition.lock();
        const debugMsg = `Creating service ${definition.name.toString()}`;
        debugCreation(`${debugMsg} - started`);
        let currentMiddleware = 0;
        const middlewares = this.getMiddlewares();
        const next = (definition) => {
            const middleware = middlewares[currentMiddleware];
            currentMiddleware++;
            if (middleware) {
                return middleware.call(this, definition, next);
            }
            else {
                return Promise.all(definition.args.map(a => a instanceof ContainerArg_1.ContainerArg ? a.getArgument(this) : a))
                    .then((args) => definition.factory.apply(this, args));
            }
        };
        const result = next.call(this, definition);
        if (isThenable(result)) {
            return result.then(x => {
                debugCreation(`${debugMsg} - finished`);
                return x;
            });
        }
        else {
            debugCreation(`${debugMsg} - finished`);
            return Promise.resolve(result);
        }
    }
    /**
     * Returns all services that definition satisfies predicate
     */
    getByPredicate(predicate) {
        return Promise.all(this.findByPredicate(predicate)
            .map(d => this.get(d)));
    }
    getByAnnotation(predicate, withAnnotation = false) {
        return Promise.all(this.findByAnnotation(predicate, true)
            .map(([definition, annotation]) => __awaiter(this, void 0, void 0, function* () {
            if (withAnnotation) {
                return [yield this.get(definition), annotation];
            }
            return this.get(definition);
        })));
    }
}
exports.Container = Container;
